/* 
 * Copyright (C) 2003 by the libdvdnav project
 * 
 * This file is part of libdvdnav, a DVD navigation library.
 * 
 * libdvdnav is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * libdvdnav is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 * $Id: menus.c,v 1.12 2003/03/29 14:19:01 mroi Exp $
 *
 */

#include <stdio.h>
#include <getopt.h>
#include <unistd.h>
#include <inttypes.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "cvdvtypes.h"
#include "dvdnav.h"
#include "dvdnav_internal.h"
#include "vm.h"

#include "dvb_formats.h"

#include <linux/dvb/video.h>
#include <linux/dvb/audio.h>

/* which is the default language for menus/audio/subpictures? */
#define DVD_MENU_LANGUAGE "en"
#define DVD_AUDIO_LANGUAGE "en"
#define DVD_SUBPICTURE_LANGUAGE "nl"
#ifdef LANG_ja
# undef DVD_MENU_LANGUAGE
# undef DVD_AUDIO_LANGUAGE
# undef DVD_SUBPICTURE_LANGUAGE
# define DVD_MENU_LANGUAGE "ja"
# define DVD_AUDIO_LANGUAGE "en"
# define DVD_SUBPICTURE_LANGUAGE "ja"
#endif /* LANG_ja */

/* shall we use libdvdnav's read ahead cache? */
#define DVD_READ_CACHE 1

//#define DEVICE_NAME "/dev/margi"
//#define DEVICE_NAME "/dev/dvb/adapter0/video0"

void DECCMD(int fd, struct decodercmd cmd)
{
	int err = ioctl(fd,_IO(CVDV_IOCTL_MAGIC,IOCTL_DECODER),&cmd);
	if (err) {
		perror("DECCMD: ioctl failed\n");
	}
}

#define  DO_IOCTL(fd, req, arg)\
	printf("%d ioctl %d\n", __LINE__, req);\
	if (ioctl(fd, req, arg)!=0) perror("DO_IOCTL: ioctl failed")

#define  DO_IOCTL0(fd, req)\
	printf("%d ioctl %d\n", __LINE__, req);\
	if (ioctl(fd, req)!=0) perror("DO_IOCTL0: ioctl failed")

#include <termios.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/extensions/Xv.h>
#include <X11/extensions/Xvlib.h>

#include <Xm/MwmUtil.h>

#define GUID_YUV12_PLANAR 0x32315659

Display *dpy;
int xv_port = -1;
Window window, root;
GC gc;

int subtitle_stream = 0;


void noborders(Display *display, Window win)
{
	MwmHints mwmhints;
	Atom MOTIF_WM_HINTS;
	mwmhints.flags = MWM_HINTS_DECORATIONS;
	mwmhints.decorations = 0;
	MOTIF_WM_HINTS = XInternAtom(display, "_MOTIF_WM_HINTS", True);
	XChangeProperty(display, win, MOTIF_WM_HINTS, MOTIF_WM_HINTS,
			sizeof(long)*8, PropModeReplace,
			(unsigned char *) &mwmhints,
			sizeof(mwmhints) / sizeof(long) );
}

void zvwin(int permitted_df)
{
	int vx, vy, vw, vh;
	int dx, dy, dw, dh;
	printf("zvwin(%d)\n", permitted_df);
	switch(permitted_df) {
	case 0:
	case 2: /* 16:9 */
		// vx = 0; vy = 0; vw = 720; vh = 576;

		vx = 16;  vy = 20; vw = 720; vh = 576;
		dx = 10;  dy = -100; dw = 800; dh = 480+140;

		//dx = 0; dy = 0; dw = 720; dh = 576;
		//dx = 0; dy = 0; dw = 720; dh = 460;
		break;
	case 1: /* 4:3 */
		//vx = 0; vy = 0; vw = 640; vh = 576;

		vx = 16;  vy = 20; vw = 720; vh = 576;
		dx = 10;  dy = -100; dw = 800; dh = 480+140;

		//dx = 40; dy = 0; dw = 640; dh = 576;
		//dx = 40; dy = 0; dw = 640; dh = 460;
		break;
	case 3: /* 4:3 */
		//vx = 0; vy = 0; vw = 640; vh = 576;

		vx = 16;  vy = 20; vw = 720; vh = 576;
		dx = 10;  dy = -100; dw = 800; dh = 480+140;
		break;
	default:
		printf("zvwin: %d should not happen!\n", permitted_df);
		break;
	}
	//0, 8, 640, 432,        // video vx,vy vw,vh 
	//0, 0, 798, 470		// drawable dx,dy dw,dh

	if (XvPutVideo(dpy, xv_port, window, gc, 
		       vx, vy, vw, vh, // video vx,vy vw,vh 
		       dx, dy, dw, dh  // drawable dx,dy dw,dh
		       ) != Success){
		printf("XvPutVideo failed.\n");
		//exit(-1);
	}
	//	noborders(dpy, window);
	XFlush(dpy);
	
}
			       
void zvinit()
{
  int screen;
  XVisualInfo vinfo;
  XSetWindowAttributes xswa;
  unsigned long mask;
  unsigned int p_version, p_release, p_request_base, p_event_base, p_error_base;
  int p_num_adaptors;
  XvAdaptorInfo *ai;
  int ret, x, y, w, h, b, d;
  //  XEvent event;

  unsigned int nr_enq;
  XvEncodingInfo *xv_enq;

  unsigned int nr_fmt;
  XvImageFormatValues *xv_fmt;

  if ((dpy = XOpenDisplay(NULL)) == NULL){
    fprintf(stderr, "Cannot open Display.\n");
    return;
  }
  screen = DefaultScreen(dpy);
  if (!XMatchVisualInfo(dpy, screen, 16, TrueColor, &vinfo)
     && !XMatchVisualInfo(dpy, screen, 15, TrueColor, &vinfo)){
    fprintf(stderr, "Cannot find the visual (15bpp or 16bpp).\n");
    exit(-1);
  }
  
  xswa.colormap =  XCreateColormap(dpy, DefaultRootWindow(dpy), vinfo.visual, AllocNone);
  xswa.event_mask = StructureNotifyMask | ExposureMask;
  xswa.background_pixel = 0;
  xswa.border_pixel = 0;
  mask = CWBackPixel | CWBorderPixel | CWColormap | CWEventMask;
  
  window = XCreateWindow(dpy, DefaultRootWindow(dpy),
			 0, 0, 800, 480, 0, vinfo.depth,
			 //30, 0, 740, 440, 0, vinfo.depth,
			 InputOutput, vinfo.visual, mask, &xswa);
  gc = XCreateGC(dpy, window, 0, 0);
  XStoreName(dpy, window, "ZV");
  XSetIconName(dpy, window, "ZV");
  XSelectInput(dpy, window, StructureNotifyMask);
  XMapWindow(dpy, window);

  XSync(dpy, 1);
  
  if ((ret = XvQueryExtension(dpy, &p_version, &p_release, &p_request_base, &p_event_base, &p_error_base)) != Success){
      fprintf(stderr, "Cannot find XV Extention.\n");
      exit(-1);
  }
  if ((ret = XvQueryAdaptors(dpy, DefaultRootWindow(dpy), &p_num_adaptors, &ai)) != Success){
      fprintf(stderr, "Cannot find XV adaptor.\n");
      exit(-1);
  }
  if (p_num_adaptors == 0){
      fprintf(stderr, "Cannot find XV adaptor.\n");
      exit(-1);
  }
  xv_port = ai[0].base_id;

  if ((ret = XvQueryEncodings(dpy, xv_port, &nr_enq, &xv_enq))  != Success) {
    fprintf(stderr, "Cannot Query Encodings\n");
    exit(-1);
  }

  if ((xv_fmt = XvListImageFormats(dpy, xv_port, &nr_fmt))  == NULL) {
    fprintf(stderr, "Cannot List Formats\n");
    exit(-1);
  }
    
  XGetGeometry(dpy, window, &root, &x, &y, &w, &h, &b, &d);
  printf("x = %d\ny = %d\nw = %d\nh = %d\nb=%d\nd =%d\n",
	 x, y, w, h, b, d);
  
}

char *dvdnav_event_types[] = {
	"DVDNAV_BLOCK_OK",
	"DVDNAV_NOP",
	"DVDNAV_STILL_FRAME",
	"DVDNAV_SPU_STREAM_CHANGE",
	"DVDNAV_AUDIO_STREAM_CHANGE",
	"DVDNAV_VTS_CHANGE",
	"DVDNAV_CELL_CHANGE",
	"DVDNAV_NAV_PACKET",
	"DVDNAV_STOP",
	"DVDNAV_HIGHLIGHT",
	"DVDNAV_SPU_CLUT_CHANGE",
	"unknown event",
	"DVDNAV_HOP_CHANNEL",
	"DVDNAV_WAIT"
};

void print_event(int event) 
{
	static int prev_event = -1;
	if ( prev_event == event) {
	} else {
		if ( (event < DVDNAV_BLOCK_OK) || 
		     (event > DVDNAV_WAIT) ) {
			printf("unknown event\n");
		} else {
			printf("event (%2d) '%s'\n", event, dvdnav_event_types[event]);
		}
		prev_event = event;
	}
}


static int use_term = 0;
static struct termios term_org;

static void term_init(void)
{
    struct termios term;

    use_term = 0;
    if (tcgetattr(0, &term) < 0){
	perror("term_init");
	return;
    }

    term_org = term;
    term.c_lflag &= ~(ICANON|ECHO); 
    term.c_cc[VMIN] = 1;
    term.c_cc[VTIME] = 0;
    if (tcsetattr(0, TCSANOW, &term) < 0){
	perror("term_init");
	return;
    }
    use_term = 1;
}

static void term_restore(void)
{
    if (use_term)
	tcsetattr(0, TCSAFLUSH, &term_org);
}

static int option_index = 0;
static struct option long_options[] = {
    {"audio", required_argument, NULL, 'a'},
    {"astream",required_argument, NULL, 'm'},
    {"subtitle",required_argument, NULL, 's'},
    {"help", no_argument , NULL, 'h'},
    {0, 0, 0, 0}
};

#if 0
char navipack[1024];
#endif

void set_subtitle(int fd, int active, int subtitle_stream)
{
	video_spu_t Aspu;
	printf("set_subtitle: stream = %d\n", subtitle_stream);
	Aspu.active = active;
	Aspu.streamID = (subtitle_stream < 0) ? 32 : subtitle_stream;

	DO_IOCTL(fd, VIDEO_SET_SPU, &Aspu);
}

void highlight_button(dvdnav_t *dvdnav, int fd, int button, int active)
{
	pci_t *pci;
	dsi_t *dsi;
	btni_t *b;
	int i;
	unsigned char coli[4], posi[6];
	video_highlight_t vh;	
	/* Applications with fifos should not use these functions to retrieve NAV packets,
	 * they should implement their own NAV handling, because the packet you get from these
	 * functions will already be ahead in the stream which can cause state inconsistencies.
	 * Applications with fifos should therefore pass the NAV packet through the fifo
	 * and decoding pipeline just like any other data. */
	printf("button %d active %d\n",
	       button, active);
	pci = dvdnav_get_current_nav_pci(dvdnav);
	dsi = dvdnav_get_current_nav_dsi(dvdnav);
	b = &(pci->hli.btnit[button - 1]);
	i = b->btn_coln - 1;
	vh.active 	= active;
	//vh.contrast1 	= coli[0] = pci->hli.btn_colit.btn_coli[i][0] >> 24;
	//vh.contrast2 	= coli[1] = pci->hli.btn_colit.btn_coli[i][0] >> 16;
	//vh.color1 	= coli[2] = pci->hli.btn_colit.btn_coli[i][0] >> 8;
	//vh.color2 	= coli[3] = pci->hli.btn_colit.btn_coli[i][0];
	vh.contrast1 	= coli[0] = pci->hli.btn_colit.btn_coli[i][0] >> 24;
	vh.contrast2 	= coli[1] = pci->hli.btn_colit.btn_coli[i][0] >> 16;
	vh.color2 	= coli[2] = pci->hli.btn_colit.btn_coli[i][0] >> 8;
	vh.color1 	= coli[3] = pci->hli.btn_colit.btn_coli[i][0];
	vh.xpos 	= vh.ypos = 0;

	//vh.ypos = (b->auto_action_mode << 22) | ((b->y_start) << 12)|(b->y_end);
	//vh.xpos = (b->btn_coln         << 22) | ((b->x_start) << 12)|(b->x_end);

	vh.xpos = 
		( ( ( ( ( b->x_start & 0x3f0) >> 4) & 0x3f) | ( (b->btn_coln << 6)      ) ) << 16) +  // xpos[0]
		( ( ( ( ( b->x_start & 0xf  ) << 4) & 0xf0) | ( (b->x_end & 0x300) >> 8 ) ) << 8 ) +  // xpos[1]
		( ( ( ( ( b->x_end   & 0xff )     )       )                               )      );   // xpos[2]

	vh.ypos =
		( ( ( ( ( b->y_start & 0x3f0) >> 4) & 0x3f) | ( (b->auto_action_mode<<6)) ) << 16) +  // ypos[0]
		( ( ( ( ( b->y_start & 0xf  ) << 4) & 0xf0) | ( (b->y_end & 0x300) >> 8 ) ) << 8 ) +  // ypos[1]
		( ( ( ( ( b->y_end   & 0xff )     )       )                               )      );   // ypos[2]

	printf("coli[0] = %3d coli[1] = %3d coli[2] = %3d coli[3] = %3d "
	       "auto_action_mode = %d btn_coln = %d "
	       "x_start = %3d x_end = %3d y_start = %3d y_end = %3d\n",
	       coli[0], coli[1], coli[2], coli[3], 
	       b->auto_action_mode, b->btn_coln, 
	       b->x_start, b->x_end,
	       b->y_start, b->y_end);
	printf("vh.contrast1 = %3d vh.contrast2 = %3d vh color1    = %3d vh.color2    = %3d "
	       "vh.xpos = %8x vh.ypos = %8x\n",
	       vh.contrast1, vh.contrast2, vh.color1, vh.color2, 
	       vh.xpos, vh.ypos);
	DO_IOCTL(fd, VIDEO_SET_HIGHLIGHT, &vh);
	//set_subtitle(fd, 1, subtitle_stream);
}


main(int argc, char **argv)
{
    int i, c;
    int written, cc;
    int len_w;
    char *audio_type = "AC3";
    int audio_stream = 0;
    char *path = NULL;
    dvdnav_t *dvdnav;
    uint8_t mem[DVD_VIDEO_LB_LEN];
    int finished = 0;
    int fd;
    int fda;
    int skip_wait = 0;
    int wait_time = 0;
    int wait_time_elapsed = 0;
   
    struct decodercmd decmd;

    video_spu_t	Aspu;
    video_spu_palette_t Aspupalette;
    video_highlight_t hl;

    unsigned char clut[3*16];
    int button;
    int menu_dir;
#define MENU_LEFT  1
#define MENU_DOWN  2
#define MENU_UP    3
#define MENU_RIGHT 4
#define MENU_ENTER 5
    fd_set rfds;
    fd_set wfds;
    struct timeval t;
    int stopped;
    int hop;
    int state;
#define STATE_NORMAL   0x00
#define STATE_NUMINPUT 0x01
#define STATE_ESC      0x02
#define STATE_ESC_BRACKET 0x04
#define STATE_AUDIO    0x10
#define STATE_TITLE    0x20
#define STATE_CHAPTER  0x30
#define STATE_MENU     0x40
#define STATE_SUBTITLE 0x50
    char keybuf[4];
    int keyinp;

    static void help(void), help2(void), finish(int);
  
    while ((c = getopt_long(argc, argv, "a:m:s:h", long_options, &option_index)) >= 0){
	switch (c){
	case 'a':
	    audio_type = optarg;
	    break;
	case 'm':
	    audio_stream = atoi(optarg);
	    break;
	case 's':
	    subtitle_stream = atoi(optarg);
	    break;
	case 'h':
	case '?':
	default:
	    help();
	    exit(0);
	}
    }

    if (optind < argc){
	path = argv[optind];
    } else {
	help();
	exit(-1);
    }


    if (open_av(&fd, &fda, 0) < 0) exit(1);
    if (audioSetStreamtype(fda, audio_type) < 0) exit(1);
    if (videoSetStreamtype(fd, "PROG") < 0) exit(1);

    if (fda) audioSetAVSync(fda, true);

    if (sizeof(int) != sizeof(boolean)) abort();

    videoSelectSource(fd, VIDEO_SOURCE_MEMORY);
    //    videoPlay(fd);

	 
    //    if ((fd = open(DEVICE_NAME, O_RDWR|O_NONBLOCK)) < 0){
    //	perror(DEVICE_NAME);
    //	exit(-1);
    //}

    zvinit();

    //decmd.cmd = Decoder_Set_Streamtype; decmd.param1 = stream_DVD; DECCMD(fd, decmd);
    //DO_IOCTL(fd, VIDEO_SET_STREAMTYPE, VIDEO_CAP_SPU);

    set_subtitle(fd, 1, subtitle_stream);

/*     decmd.cmd = Decoder_SPU; */
/*     decmd.param1 = (subtitle_stream < 0) ? 32 : subtitle_stream; */
/*     decmd.param2 = 1; */
/*     DECCMD(fd, decmd); */

    /* open dvdnav handle */
    if (dvdnav_open(&dvdnav, path) != DVDNAV_STATUS_OK) {
	fprintf(stderr, "Error: dvdnav_open %s\n", path);
	exit(1);
    }
#if 0
    dvdnav_set_region_mask(dvdnav, 0x02);
#endif

    /* set read ahead cache usage */
    if (dvdnav_set_readahead_flag(dvdnav, DVD_READ_CACHE) != DVDNAV_STATUS_OK) {
	fprintf(stderr, "Error: dvdnav_set_readahead_flag: %s\n", dvdnav_err_to_string(dvdnav));
	exit(2);
    }
  
    /* set the language */
    if (dvdnav_menu_language_select(dvdnav, DVD_MENU_LANGUAGE) != DVDNAV_STATUS_OK ||
	dvdnav_audio_language_select(dvdnav, DVD_AUDIO_LANGUAGE) != DVDNAV_STATUS_OK ||
	dvdnav_spu_language_select(dvdnav, DVD_SUBPICTURE_LANGUAGE) != DVDNAV_STATUS_OK) {
	fprintf(stderr, "Error: setting languages: %s\n", dvdnav_err_to_string(dvdnav));
	exit(2);
    }
  
    /* set the PGC positioning flag to have position information relatively to the
     * whole feature instead of just relatively to the current chapter */
    if (dvdnav_set_PGC_positioning_flag(dvdnav, 1) != DVDNAV_STATUS_OK) {
	fprintf(stderr, "Error: dvdnav_set_PGC_positioning_flag: %s\n", dvdnav_err_to_string(dvdnav));
	exit(2);
    }
  
    term_init();
    FD_ZERO(&rfds);
    state = STATE_NORMAL;
    stopped = 0;
    hop = 0;

    /* the read loop which regularly calls dvdnav_get_next_block
     * and handles the returned events */
    while (!finished) {
	int result, event, len, ret;
	uint8_t *buf = mem;

	if (use_term){
	    FD_SET(STDIN_FILENO, &rfds);
	    t.tv_sec = wait_time;
	    t.tv_usec = 0;
	    if (select(STDIN_FILENO+1, &rfds, NULL, NULL, &t) == 1){
		do{
		    read(STDIN_FILENO, &c, 1);
		    c &= 0xff;
		    if ((state & 0x0f) == STATE_NORMAL){
			switch (c){
			case 'w':
#if 1
/* 			    decmd.cmd = Decoder_AVSYNC; */
/* 			    decmd.param1 = 2; */
/* 			    decmd.param2 = 30000; */
/* 			    DECCMD(fd, decmd); */
				DO_IOCTL(fda, AUDIO_SET_AV_SYNC, true);
#endif
			    break;
			case '?':
			    help2();
			    break;
			case 'i': /* info */
			    {
				int n_title, title, part, n_part;
				dvdnav_current_title_info(dvdnav, &title, &part);
				dvdnav_get_number_of_parts(dvdnav, title, &n_part);
				printf("[Current] Title: %2d Chapter: %2d\n", 
				       title, part, n_part);
				dvdnav_get_number_of_titles(dvdnav, &n_title);
				for (title = 1; title <= n_title; title++){
				    dvdnav_get_number_of_parts(dvdnav, title, &n_part);
				    printf("  Title %2d: %2d chapters\n", 
					   title, n_part);
				}
			    }
			    break;
			case ' ': /* pause toggle */
			    stopped = !stopped;
			    if (stopped) {
				    printf("play\n");
				    DO_IOCTL0(fd, VIDEO_CONTINUE);
			    } else {
				    printf("pauze\n");
				    DO_IOCTL0(fd, VIDEO_FREEZE);
			    }
			    break;
			case 't':
			    printf("title menu\n");
			    if (dvdnav_menu_call(dvdnav, DVD_MENU_Title) == DVDNAV_STATUS_OK)
				hop = 1;
			    state = STATE_NORMAL;
			    break;
			case 'r':
			    printf("root menu\n");
			    if (dvdnav_menu_call(dvdnav, DVD_MENU_Root) == DVDNAV_STATUS_OK)
				hop = 1;
			    state = STATE_NORMAL;
			    break;
#if 0
			case 'n': /* next chapter */
			case 'p': /* prev chapter */
			    if (state == STATE_NORMAL){
				int title, part, n_part, new;
				dvdnav_current_title_info(dvdnav, &title, &part);
				dvdnav_get_number_of_parts(dvdnav, title, &n_part);
				printf("title: %2d  chapter: %2d (last=%d)\n", title, part, n_part);
				new = -1;
				if (c == 'n'){
				    if (part < n_part)
					new = part + 1;
				} else {
				    if (part > 1)
					new = part - 1;
				}
				if (new > 0)
				    dvdnav_part_play(dvdnav, title, new);
			    }
			    break;
#else
			case 'n':
				dvdnav_next_pg_search(dvdnav);
				break;
			case 'p':
				dvdnav_prev_pg_search(dvdnav);
				break;
#endif
			case 'T': /* title */
			    if (state == STATE_NORMAL){
				state = STATE_TITLE|STATE_NUMINPUT;
				keyinp = 0;
			    }
			    break;
			case 'C': /* chapter */
			    if (state == STATE_NORMAL){
				state = STATE_CHAPTER|STATE_NUMINPUT;
				keyinp = 0;
			    }
			    break;
			case 'a': /* select audio track */
			    if (state == STATE_NORMAL){
				state = STATE_AUDIO|STATE_NUMINPUT;
				keyinp = 0;
			    }
			    break;
			case 's': /* select subtitle stream */
			    if (state == STATE_NORMAL){
				state = STATE_SUBTITLE|STATE_NUMINPUT;
				keyinp = 0;
			    }
			    break;
			case 'h':
			    menu_dir = MENU_LEFT;
			    break;
			case 'j':
			    menu_dir = MENU_DOWN;
			    break;
			case 'k':
			    menu_dir = MENU_UP;
			    break;
			case 'l':
			    menu_dir = MENU_RIGHT;
			    break;
			case '\n': case '\r':
			    menu_dir = MENU_ENTER;
			    skip_wait = 1;
			    break;
			case 'q':
			    printf("quit\n");
			    finished = 1;
			    break;
			case '\033':
			    state |= STATE_ESC;
			    break;
			}
		    } else if (state & STATE_ESC){
			state &= 0xf0;
			if (c == '[')
			    state |= STATE_ESC_BRACKET;
		    } else if (state & STATE_ESC_BRACKET){
			state &= 0xf0;
			switch (c){
			case 'A':
			    menu_dir = MENU_UP;
			    if (state != STATE_MENU){
			        dvdnav_sector_search(dvdnav, -15000, SEEK_CUR);
			    }
			    break;
			case 'B':
			    menu_dir = MENU_DOWN;
			    if (state != STATE_MENU){
			        dvdnav_sector_search(dvdnav, 15000, SEEK_CUR);
			    }
			    break;
			case 'C':
			    menu_dir = MENU_RIGHT;
			    if (state != STATE_MENU){
			        dvdnav_sector_search(dvdnav, 5000, SEEK_CUR);
			    }
			    break;
			case 'D':
			    menu_dir = MENU_LEFT;
			    if (state != STATE_MENU){
			        dvdnav_sector_search(dvdnav, -5000, SEEK_CUR);
			    }
			    break;
			}
		    } else if (state & STATE_NUMINPUT){
			keybuf[keyinp] = c;
			if (c == '\r' || c == '\n'){
			    keybuf[keyinp] = '\0';
			    c = atoi(keybuf);
			    switch (state & 0xf0){
			    case STATE_AUDIO:
				audio_stream = c;
				printf("Audio stream: %d\n", audio_stream);
				DO_IOCTL(fda, AUDIO_SET_ID, audio_stream);
/* 				decmd.cmd = Decoder_Set_AudioStreamID; */
/* 				decmd.param1 = audio_stream & 0x07; */
/* 				decmd.param2 = 0x10|(audio_stream & 0x07); */
/* 				DECCMD(fd, decmd); */
			    	state = STATE_NORMAL;
				break;
			    case STATE_SUBTITLE:
				subtitle_stream = c;
				printf("Subtitle stream: %d\n", subtitle_stream);
				set_subtitle(fd, 1, subtitle_stream);
			    	state = STATE_NORMAL;
				break;
			    case STATE_TITLE:
			        {
				    int n_title;
				    dvdnav_get_number_of_titles(dvdnav, &n_title);
				    printf("TITLE: %d (last=%d)\n", c, n_title);
				    if (c > 0  &&  c <= n_title){
					dvdnav_title_play(dvdnav, c);
				    }
			        }
			    	state = STATE_NORMAL;
				break;
			    case STATE_CHAPTER:
			        {
				    int title, part, n_part, new;
				    dvdnav_current_title_info(dvdnav, &title, &part);
				    dvdnav_get_number_of_parts(dvdnav, title, &n_part);
				    printf("CHAPTER: %d (last=%d)\n", c, n_part);
				    if (c > 0  &&  c <= n_part){
					dvdnav_part_play(dvdnav, title, c);
				    }
			        }
			    	state = STATE_NORMAL;
				break;
			    }
			} else {
			    keyinp++;
			    if (keyinp == sizeof keybuf)
				state = STATE_NORMAL;
			}
		    }
		} while (stopped  &&  !finished);
	    } else {
		    /* timeout expired */
		    if (wait_time) {
			    wait_time_elapsed = 1;
		    }
	    }
	}
	if (finished)
	    continue;

#if 0
	decmd.cmd = Decoder_GetNavi;
	decmd.data1 = navipack;
	if ((result = DECCMD(fd, decmd)) < 0){
	    fprintf(stderr, "NAVIPACK Fail\n");
	} else if (result > 0){
	    fprintf(stderr, "NAVIPACK %d\n", result);
	}
#endif
    
	/* the main reading function */
#if DVD_READ_CACHE
	result = dvdnav_get_next_cache_block(dvdnav, &buf, &event, &len);
#else
	result = dvdnav_get_next_block(dvdnav, buf, &event, &len);
#endif
	if (result == DVDNAV_STATUS_ERR) {
	    fprintf(stderr, "Error: getting next block: %s\n", dvdnav_err_to_string(dvdnav));
	    finish(3);
	}

	print_event(event);

	switch (event){
	case DVDNAV_BLOCK_OK:
		{
			char *bufp = buf;
			written = 0;
			while(len > 0 && written<len ) {
				FD_ZERO(&wfds);
				FD_SET(fd, &wfds);
				t.tv_sec = 0;
				t.tv_usec = 500000;
				if (select(fd+1, NULL, &wfds, NULL, &t) == 1) {
					len_w = write(fd, buf+written, len-written);
					if (len_w > 0) written += len_w;
				} else {
					perror("select time out");
					fprintf(stderr, 
						"len     %d\n"
						"len_w   %d\n"
						"written %d\n",
						len, len_w, written);
					break;
				}
			}
		}
	    break;
	case DVDNAV_NOP:
	    /* Nothing to do here. */
	    break;
	case DVDNAV_STILL_FRAME: 
	    /* We have reached a still frame. A real player application would wait
	     * the amount of time specified by the still's length while still handling
	     * user input to make menus and other interactive stills work.
	     * A length of 0xff means an indefinite still which has to be skipped
	     * indirectly by some user interaction. */
	    {
		dvdnav_still_event_t *still_event = (dvdnav_still_event_t *)buf;
		if (still_event->length < 0xff) {
			printf("Skipping %d seconds of still frame\n", still_event->length);
			wait_time = still_event->length;
			wait_time_elapsed = 0;
		} else {
			//printf("Skipping indefinite length still frame\n");
			wait_time = 0;
		}
		if (wait_time_elapsed) {
			dvdnav_still_skip(dvdnav);
			wait_time_elapsed = 0;
		}
	    }
	    break;
	case DVDNAV_WAIT:
	    /* We have reached a point in DVD playback, where timing is critical.
	     * Player application with internal fifos can introduce state
	     * inconsistencies, because libdvdnav is always the fifo's length
	     * ahead in the stream compared to what the application sees.
	     * Such applications should wait until their fifos are empty
	     * when they receive this type of event. */
	    // sleep(1);
	    if (skip_wait) {
		printf("Skipping wait condition\n");
		dvdnav_wait_skip(dvdnav);
		skip_wait = 0;
	    }
	    break;
	case DVDNAV_SPU_CLUT_CHANGE:
	    /* Player applications should pass the new colour lookup table to their
	     * SPU decoder */
	    for (i = 0; i < len/4; i++){

		clut[i*3+0] = buf[i*4+2];
		clut[i*3+1] = buf[i*4+1];
		clut[i*3+2] = buf[i*4+0];
	    }

	    Aspupalette.palette = clut;
	    Aspupalette.length  = len / 4 * 3;
	    DO_IOCTL(fd, VIDEO_SET_SPU_PALETTE, &Aspupalette);
	    break;
	case DVDNAV_SPU_STREAM_CHANGE:
	    /* Player applications should inform their SPU decoder to switch channels */
	    {
		dvdnav_spu_stream_change_event_t *spu_ev;

	        spu_ev = (dvdnav_spu_stream_change_event_t *)buf;
	        fprintf(stderr, "DVDNAV_SPU_STREAM_CHANGE: %d %d %d %d\n",
			spu_ev->physical_wide, spu_ev->physical_letterbox,
			spu_ev->physical_pan_scan, spu_ev->logical);
	        if (subtitle_stream >= 0  &&  spu_ev->physical_letterbox >= 0  
		    &&  spu_ev->physical_letterbox != subtitle_stream){
			//subtitle_stream = spu_ev->physical_letterbox;

		    set_subtitle(fd, 1, subtitle_stream);
	        }
	    }
	    break;
	case DVDNAV_AUDIO_STREAM_CHANGE:
	    /* Player; applications should inform their audio decoder to switch channels */
	    {
		dvdnav_audio_stream_change_event_t *audio_ev;

	        audio_ev = (dvdnav_audio_stream_change_event_t *)buf;
	        fprintf(stderr, "DVDNAV_AUDIO_STREAM_CHANGE: %d %d\n", audio_ev->physical, audio_ev->logical);
	        if (audio_ev->physical >= 0  &&  audio_ev->physical != audio_stream){
			DO_IOCTL(fda, AUDIO_SET_ID, audio_stream);
	        }
	    }
	    break;
	case DVDNAV_HIGHLIGHT:
	    /* Player applications should inform their overlay engine to highlight the 
	     * given button */
	    {
		dvdnav_highlight_event_t *highlight_event = (dvdnav_highlight_event_t *)buf;
		pci_t *pci;
		dsi_t *dsi;
		btni_t *b;
		printf("Selected button %d\n", highlight_event->buttonN);
		button = highlight_event->buttonN;
		highlight_button(dvdnav, fd, button, 1);
	    }
	    break;
	case DVDNAV_VTS_CHANGE:
	    /* Some status information like video aspect and video scale permissions do
	     * not change inside a VTS. Therefore this event can be used to query such
	     * information only when necessary and update the decoding/displaying
	     * accordingly. */
	    {
		dvdnav_vts_change_event_t *vts_ev;
		video_attr_t v;
		unsigned short attr;
		unsigned int flush =1;

		vts_ev = (dvdnav_vts_change_event_t *)buf;
	        fprintf(stderr, "DVDNAV_VTS_CHANGE: %d->%d\n", vts_ev->old_vtsN, vts_ev->new_vtsN);
	        if (vts_ev->new_vtsN < 0)
		    break;

		v = vm_get_video_attr(dvdnav->vm);
		attr = (v.mpeg_version<<14)|(v.video_format<<12)|(v.display_aspect_ratio<<10)|(v.permitted_df<<8)|(v.line21_cc_1<<7)|(v.line21_cc_2<<6)|(v.picture_size<<3)|(v.letterboxed<<2)|(v.film_mode<<0);
		printf("Video attibutes %08x\n", attr);
		//DO_IOCTL(fd, VIDEO_SET_ATTRIBUTES, attr);
		zvwin(v.permitted_df);
		DO_IOCTL0(fd, VIDEO_CONTINUE);
	    }
	    break;
	case DVDNAV_CELL_CHANGE:
	    /* Some status information like the current Title and Part numbers do not
	     * change inside a cell. Therefore this event can be used to query such
	     * information only when necessary and update the decoding/displaying
	     * accordingly. */
	    {
		int tt = 0, ptt = 0, pos, len;

		dvdnav_current_title_info(dvdnav, &tt, &ptt);
		dvdnav_get_position(dvdnav, &pos, &len);
		fprintf(stderr, "Cell change: Title %d, Chapter %d\n", tt, ptt);
		fprintf(stderr, "At position %.0f%% inside the feature\n", 100 * (double)pos / (double)len);
	    }
	    break;
	case DVDNAV_NAV_PACKET:
	    /* A NAV packet provides PTS discontinuity information, angle linking information and
	     * button definitions for DVD menus. Angles are handled completely inside libdvdnav.
	     * For the menus to work, the NAV packet information has to be passed to the overlay
	     * engine of the player so that it knows the dimensions of the button areas. */
	    {
		pci_t *pci;
		dsi_t *dsi;
		btni_t *b;
		unsigned char coli[4], posi[6];
		video_highlight_t vh;	
		/* Applications with fifos should not use these functions to retrieve NAV packets,
		 * they should implement their own NAV handling, because the packet you get from these
		 * functions will already be ahead in the stream which can cause state inconsistencies.
		 * Applications with fifos should therefore pass the NAV packet through the fifo
		 * and decoding pipeline just like any other data. */
		pci = dvdnav_get_current_nav_pci(dvdnav);
		dsi = dvdnav_get_current_nav_dsi(dvdnav);
	
		if(pci->hli.hl_gi.btn_ns > 0) {
	
		    if (state == STATE_NORMAL){
		        printf("Found %i DVD menu buttons...\n", pci->hli.hl_gi.btn_ns);
		        for (i = 0; i < pci->hli.hl_gi.btn_ns; i++){
			    b = &(pci->hli.btnit[i]);
			    printf("Button %i top-left @ (%i,%i), bottom-right @ (%i,%i)\n", 
			           i+1, b->x_start, b->y_start, b->x_end, b->y_end);
		        }
			state = STATE_MENU;
			menu_dir = 0;
		    }
		    /* validate button value */
		    if (button <= 0)
			button = 1;
		    else if (button > pci->hli.hl_gi.btn_ns)
			button = pci->hli.hl_gi.btn_ns;
		    b = &(pci->hli.btnit[button - 1]);
		    switch (menu_dir){
		    case MENU_LEFT:
			button = b->left;
			menu_dir = 0;
			break;
		    case MENU_DOWN:
			button = b->down;
			menu_dir = 0;
			break;
		    case MENU_UP:
			button = b->up;
			menu_dir = 0;
			break;
		    case MENU_RIGHT:
			button = b->right;
			menu_dir = 0;
			break;
		    }

		    highlight_button(dvdnav, fd, button, 1);

		    if (menu_dir == MENU_ENTER){
			state = STATE_NORMAL;
			menu_dir = 0;
			vh.active = 0;
			highlight_button(dvdnav, fd, button, 0);

			printf("Selecting button %i...\n", button);
			/* This is the point where applications with fifos have to hand in a NAV packet
			 * which has traveled through the fifos. See the notes above. */
			dvdnav_button_select_and_activate(dvdnav, pci, button);
			hop = 2;
		    }
		}
	    }
	    break;
	case DVDNAV_HOP_CHANNEL:
	    /* This event is issued whenever a non-seamless operation has been executed.
	     * Applications with fifos should drop the fifos content to speed up responsiveness. */
	    fprintf(stderr, "DVDNAV_HOP_CHANNEL: %d\n", hop);
	    if (hop == 1){
		    DO_IOCTL(fd, VIDEO_DECODER_FLUSH, 1); /* discard */
#if 0
	    } else if (hop == 2){
#else
	    } else {
#endif
		    DO_IOCTL(fd, VIDEO_DECODER_FLUSH, 0);
	    }
	    hop = 0;
	    break;
	case DVDNAV_STOP:
	    /* Playback should end here. */
	    {
		finished = 1;
	    }
	    break;
	default:
	    fprintf(stderr, "Unknown event (%i)\n", event);
	    finished = 1;
	    break;
	}
#if DVD_READ_CACHE
	dvdnav_free_cache_block(dvdnav, buf);
#endif
    }
  
    /* destroy dvdnav handle */
    if (dvdnav_close(dvdnav) != DVDNAV_STATUS_OK) {
	fprintf(stderr, "Error: dvdnav_close: %s\n", dvdnav_err_to_string(dvdnav));
	finish(5);
    }
    close(fd);
  
    finish(0);
    }

static void finish(int code)
{
	if (dpy) XCloseDisplay(dpy);
	term_restore();
	exit(code);
}

static void help(void)
{
    fprintf (stderr, "Usage: margi-dvd [options] {device|directory}\n");
    fprintf (stderr, "options:\n");
    fprintf (stderr, "  --help,    -h:  print help message\n");  
    fprintf (stderr, "  --audio,   -a:  set audio type (AC3, LPCM, MPEG)\n");
    fprintf (stderr, "  --subtitle,-s:  set subtitle stream number (default:-1)\n");
}

static void help2(void)
{
    printf("?           : print this message\n");
    printf("q           : exit\n");
    printf("i           : print title information\n");
    printf("t           : title menu\n");
    printf("r           : root menu\n");
    printf("aN          : audio stream N\n");
    printf("sN          : subtitle stream N\n");
    printf("TN          : title N\n");
    printf("CN          : chapter N\n");
    printf("n           : next chapter\n");
    printf("p           : prev chapter\n");
    printf("SPACE       : pause (toggle)\n");
    printf("CURSOR left : skip backword, menu select left\n");
    printf("CURSOR down : skip backword, menu select down\n");
    printf("CURSOR up   : skip forward,  menu select up\n");
    printf("CURSOR right: skip forward,  menu select right\n");
    printf("h           : menu select left\n");
    printf("j           : menu select down\n");
    printf("k           : menu select up\n");
    printf("l           : menu select right\n");
}
